import json
import sys
from pathlib import Path

import click
import openai
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.theme import Theme

from . import gpt_integration as openai_utils
from .templates import handle_template

BLUE = "\x1b[34m"
RED = "\x1b[91m"
RESET = "\x1b[0m"

DEFAULT_MODEL = "gpt-4o-mini"


def prepare_and_generate_response(
    system: str,
    template: str,
    model: str,
    emoji: bool,
    prompt_input: str,
    temperature: float,
    tokens: bool,
    no_stream: bool,
    raw: bool,
    debug: bool,
):
    """
    Handles the parameters.
    """
    if not system:
        system = ""

    if template:
        system, prompt_input, template_model = handle_template(
            template, system, prompt_input, model
        )
        # If a model name is given in the options,
        # it will bypass the model name in the template.
        if model == DEFAULT_MODEL:
            model = template_model

    if emoji:
        system = add_emoji(system)

    prompt = openai_utils.format_prompt(system, prompt_input)

    # Temporary reformatting of the prompt for `o1` models
    # as they don't support system messages yet.
    if "o1" in model:
        prompt = [
            {
                "role": "user",
                "content": prompt_input,
            },
        ]

    if debug:
        display_debug_information(prompt, model, temperature)

    if tokens:
        display_tokens_count_and_cost(prompt, model)

    stream = not no_stream

    content, response_time, response = generate_response(
        model,
        prompt,
        raw,
        stream,
        temperature,
    )

    return content, response_time, response


def add_emoji(system: str) -> str:
    """
    Adds an emoji to the system message.
    """
    emoji_message = (
        "Add plenty of emojis as a colorful way to convey emotions. However, don't mention it."
    )
    system = system.rstrip()

    if system == "":
        return emoji_message

    if not system.endswith("."):
        system += "."
    return system + " " + emoji_message


def get_config_path() -> Path:
    """
    Gets the path to the config file.
    """
    config_path = Path.home() / ".config/lmt/config.json"
    return config_path


def load_config() -> dict:
    """
    Loads the config file.
    """
    config_path = get_config_path()
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.touch(exist_ok=True)

    try:
        with open(config_path, "r", encoding="UTF-8") as file:
            config = json.load(file)
    except json.decoder.JSONDecodeError:
        config = {}

    return config


def save_config(config: dict) -> None:
    """
    Saves the config file.
    """
    config_path = get_config_path()
    with open(config_path, "w", encoding="UTF-8") as config_path:
        json.dump(config, config_path, indent=4)


def get_markdown_code_block_theme() -> str:
    """
    Gets the markdown code block theme from the config file.
    """
    config = load_config()
    config.setdefault("code_block_theme", "monokai")
    save_config(config)

    return config["code_block_theme"]


def get_markdown_inline_code_theme() -> str:
    """
    Gets the markdown inline code theme from the config file.
    """
    config = load_config()
    config.setdefault("inline_code_theme", "blue on black")
    save_config(config)

    return config["inline_code_theme"]


def generate_response(
    model: str = DEFAULT_MODEL,
    prompt: str = None,
    raw: bool = False,
    stream: bool = True,
    temperature: float = 1,
):
    """
    Generates a response from a ChatGPT.
    """
    openai.api_key = get_api_key()

    if not openai.api_key:
        click.echo(f"{click.style('Error:', fg='red')} You need to set your OpenAI API key.")
        click.echo("You can do so by running:", nl=False)
        click.echo(f"  {click.style('lmt key set', fg='blue')}\n")
        sys.exit(1)

    # Theming for Rich Markdown
    code_block_theme = get_markdown_code_block_theme()
    inline_code_theme = get_markdown_inline_code_theme()
    custom_theme = Theme({"markdown.code": inline_code_theme})

    console = Console(theme=custom_theme)
    markdown_stream = ""
    with Live(markdown_stream, console=console, refresh_per_second=25) as live:
        # Allows rich markdown formatted stream in real time
        def update_markdown_stream(chunk: str) -> None:
            nonlocal markdown_stream
            markdown_stream += chunk
            if raw:
                print(chunk, end="")
            else:
                rich_markdown_stream = Markdown(
                    markdown_stream,
                    code_theme=code_block_theme,
                )
                live.update(rich_markdown_stream)

        try:
            content, response_time, response = openai_utils.chatgpt_request(
                api_key=openai.api_key,
                prompt=prompt,
                model=model,
                stream=stream,
                temperature=temperature,
                update_markdown_stream=update_markdown_stream,
            )

            # This is temporary to ensure that the last line always ends with a newline
            # This will be removed when refactored
            if not content.endswith("\n"):
                content += "\n"
            #############################

            if not stream:
                print(content)

        except openai.error.RateLimitError as error:
            click.echo(f"{RED}Error:{RESET} {error}", err=True)
            openai_utils.handle_rate_limit_error()
            sys.exit(1)

        except openai.error.AuthenticationError:
            openai_utils.handle_authentication_error()
            sys.stderr.write("\nYou can set your API key by running: ")
            sys.stderr.write(f"{BLUE}lmt key set{RESET}\n")
            sys.exit(1)

        except Exception as error:
            click.echo(f"{RED}Error:{RESET} {error}", err=True)
            sys.exit(1)

        else:
            return content, response_time, response


def display_debug_information(prompt, model, temperature):
    """
    Displays debug information.
    """
    click.echo("---\n" + click.style("Debug information:", fg="yellow"), err=True)
    click.echo(err=True)

    click.echo(click.style("Prompt:", fg="red"), nl=False, err=True)
    for role in prompt:
        click.echo(err=True)
        click.echo(click.style(f"{role['role']}:", fg="blue"), err=True)
        click.echo(f"{role}", err=True)
    click.echo(err=True)

    click.echo(click.style("Model:", fg="red"), err=True)
    click.echo(f"{model=}", err=True)
    click.echo(err=True)

    click.echo(click.style("Temperature:", fg="red"), err=True)
    click.echo(f"{temperature=}", err=True)
    click.echo(err=True)

    click.echo(click.style("End of debug information.", fg="yellow"), err=True)
    click.echo("---\n", err=True)


def display_tokens_count_and_cost(prompt, model):
    """
    Displays the number of tokens in the prompt and the cost of the prompt.
    """
    # If the model is of the `o1` variant, it will ignore the system message.
    # This is a temporary solution until the `o1` models support system messages.
    if "o1" in model:
        full_prompt = prompt[0]["content"]
    else:
        full_prompt = prompt[0]["content"] + prompt[1]["content"]

    # The model name `chatgpt-4o-latest` is not found in the `tiktoken` OpenAI API (as of 2024-08-14).
    # Since it's a `gpt-4o` variant anyway, we can just use `gpt-4o`.
    if model == "chatgpt-4o-latest":
        model = "gpt-4o"

    number_of_tokens = openai_utils.num_tokens_from_string(full_prompt, model)
    cost = openai_utils.estimate_prompt_cost(prompt, model)

    click.echo(
        f"Number of tokens in the prompt: {click.style(str(number_of_tokens), fg='yellow')}."
    )
    click.echo(
        f"Cost of the prompt for the {click.style(model, fg='blue')} model is:"
        f" {click.style(f'${cost}', fg='yellow')}."
    )
    click.echo(
        "Please note that this cost applies only to the prompt, not the subsequent response."
    )
    sys.exit(0)


def get_api_key() -> str:
    """
    Return the OpenAI API key.
    """
    key_file_path = get_api_key_path()
    with open(key_file_path, "r", encoding="UTF-8") as key_file:
        return key_file.read().strip()


def get_api_key_path() -> Path:
    """
    Return the path to the keys file.
    """
    key_file_path = Path.home() / ".config" / "lmt" / "key.env"
    if not key_file_path.exists():
        key_file_path.parent.mkdir(parents=True, exist_ok=True)
        key_file_path.touch()
    return key_file_path


def write_key(key: str) -> None:
    """
    Write the OpenAI API key to the key file.
    """
    key_file_path = get_api_key_path()
    with open(key_file_path, "w", encoding="UTF-8") as key_file:
        key_file.write(key)


def set_key() -> None:
    """
    Add the OpenAI API key.
    """
    key_path = get_api_key_path()
    key = get_api_key()
    if key:
        click.echo(click.style("Error: ", fg="red") + "API key already exists.")
        click.echo(f"Use `{click.style('lmt key edit', fg='blue')}` to edit it.")
        return

    key = click.prompt("Your OpenAI API key", hide_input=True)
    write_key(key)
    click.echo(f"{click.style('Success!', fg='green')} API key added.")
    click.echo(f"\nThe API key is stored in {key_path}.")


def edit_key() -> None:
    """
    Edit the OpenAI API key.
    """
    key_file_path = get_api_key_path()
    key = get_api_key()
    if not key:
        click.echo(click.style("Error: ", fg="red") + "API key does not exist.")
        click.echo("You will now be prompted to add it.\n")
        set_key()
        return

    original_key = key
    new_key = click.prompt("Your OpenAI API key", hide_input=True)
    if original_key != new_key:
        write_key(new_key)
        click.echo(f"{click.style('Success!', fg='green')} API key was updated.")
    else:
        click.echo("No changes were made.")
    click.echo(f"\nThe API key is stored in {key_file_path}.")



import sys
from pathlib import Path

import click
import yaml


def handle_template(template: str, system: str, prompt_input: str, model: str) -> tuple:
    """
    Handles the template used for the prompt.
    """
    template_content = get_template_content(template)
    system = update_from_template(template_content, "system", system)
    prompt_input = update_from_template(template_content, "user", prompt_input)
    model_template = template_content.get("model", model) or model

    return system, prompt_input, model_template


def update_from_template(template_content, key, value):
    """
    Updates the value of a key from a template.
    """
    existing_value = template_content.setdefault(key, "")
    if existing_value is None:
        template_content[key] = existing_value = ""

    return existing_value.rstrip() + (value or "")


def get_template_content(template):
    """
    Reads the template YAML file and returns a dictionary with its content.
    """
    template_file = TEMPLATES_DIR / f"{template}.yaml"

    try:
        with open(template_file, "rt", encoding="UTF-8") as file:
            template_content = yaml.safe_load(file)
    except FileNotFoundError:
        click.echo(
            click.style("Error: ", fg="red")
            + f"The template '{click.style(template, fg='red')}' does not exist."
        )
        sys.exit(1)
    else:
        return template_content


def get_templates_dir() -> Path:
    """
    Returns the path to the templates directory.
    """
    templates_dir = Path.home() / ".config" / "lmt" / "templates"
    templates_dir.mkdir(parents=True, exist_ok=True)
    return templates_dir


def get_default_template_file_path() -> Path:
    """
    Returns the path to the default template file.
    """
    default_dir = Path.home() / ".config" / "lmt" / "default"
    default_dir.mkdir(parents=True, exist_ok=True)

    default_template_file = default_dir / "template.yaml"
    if not default_template_file.exists():
        click.echo("The default template does not exist. Creating it...")
        with open(default_template_file, "w", encoding="UTF-8") as file:
            file.write(DEFAULT_TEMPLATE_CONTENT)

    return default_template_file


TEMPLATES_DIR = get_templates_dir()

DEFAULT_TEMPLATE_CONTENT = """# Documentation: https://github.com/sderev/lmt

# You may leave either of the fields empty. 


# Here, you can instruct how you want ChatGPT to behave.
# For example, you might say:
# "You are an AI modeled after Emil Cioran, the Romanian philosopher and essayist.
# You have a deep understanding of existentialism and philosophical pessimism."
# The more precise and detailed, the better!
system:




# Here, you should write what you want to say to ChatGPT.
# For example: "As a language model yourself, how do you view your own existence?"
user:




# You can change the model according to the list below.
# The default model is "gpt-3.5-turbo".
# More advanced models might provide more accurate or detailed responses,
# but they also have a higher "cost per 1K tokens". Tokens refer to chunks of text that AI models read. More tokens usually mean more cost.
model: "gpt-3.5-turbo"


# Valid Models
#
# "chatgpt" or "gpt-3.5-turbo"
# "chatgpt-16k" or "gpt-3.5-turbo-16k"
#
# "3.5" or "gpt-3.5-turbo"
# "3.5-16k" or "gpt-3.5-turbo-16k"
#
# "4" or "gpt-4"
# "gpt4" or "gpt-4"
#
# "4-32k" or "gpt-4-32k"
# "gpt4-32k" or "gpt-4-32k"
"""



import filecmp
import shutil
import sys

import click
from click_default_group import DefaultGroup

from .lib import DEFAULT_MODEL, edit_key, prepare_and_generate_response, set_key
from .templates import TEMPLATES_DIR, get_default_template_file_path

VALID_MODELS = {
    "gpt-3.5-turbo": (
        "chatgpt",
        "3.5",
    ),
    "gpt-3.5-turbo-instruct": None,
    "gpt-4": (
        "4",
        "gpt4",
    ),
    "gpt-4-turbo": (
        "4t",
        "4-turbo",
        "gpt4-turbo",
    ),
    "gpt-4-32k": (
        "4-32k",
        "gpt4-32k",
    ),
    "gpt-4o": ("4o",),
    "gpt-4o-2024-05-13": None,
    "gpt-4o-2024-08-06": None,
    "gpt-4o-2024-11-20": None,
    "gpt-4o-mini": (
        "4o-mini",
        "4omini",
        "4om",
    ),
    "gpt-4o-mini-2024-07-18": None,
    "chatgpt-4o-latest": None,
    "o1": None,
    "o1-2024-12-17": None,
    "o1-preview": None,
    "o1-preview-2024-09-12": None,
    "o1-mini": None,
    "o1-mini-2024-09-12": None,
    "gpt-4.1": ("4.1",),
    "gpt-4.1-2025-04-14": None,
    "gpt-4.1-mini": ("4.1-mini",),
    "gpt-4.1-mini-2025-04-14": None,
    "gpt-4.1-nano": ("4.1-nano",),
    "gpt-4.1-nano-2025-04-14": None,
    "gpt-4.5-preview": None,
    "o3": None,
    "o3-2025-04-16": None,
    "o3-mini": None,
    "o3-mini-2025-01-31": None,
    "o4-mini": None,
    "o4-mini-2025-04-16": None,
}


# The first two parameters are required by Click for a callback.
def validate_model_name(ctx, param, value):
    """
    Validates the model name parameter.
    """
    # This is the value that the user entered for the model name.
    model_name = value.lower()

    for model, aliases in VALID_MODELS.items():
        if model_name == model:
            return model
        if aliases is not None and model_name in aliases:
            return model

    error_message = (
        f"{click.style('Invalid model name.', fg='red')}\n"
        f"{click.style('To see the model names and their aliases, use:', fg='blue')} lmt models"
    )

    raise click.BadParameter(error_message)


# The first two parameters are required by Click for a callback.
def validate_temperature(ctx, param, value):
    """
    Validates the temperature parameter.
    """
    if 0 <= value <= 2:
        return value

    raise click.BadParameter("Temperature must be between 0 and 2.")


@click.group(cls=DefaultGroup, default="prompt", default_if_no_args=True)
@click.version_option(package_name="lmterminal")
def lmt():
    """
    Talk to ChatGPT.

    Documentation: https://github.com/sderev/lmt
    """


@lmt.command()
@click.argument(
    "prompt_input",
    type=str,
    required=False,
    nargs=-1,
)
@click.option(
    "--model",
    "-m",
    default=DEFAULT_MODEL,
    help="The model to use for the requests.",
    callback=validate_model_name,
)
@click.option(
    "--template",
    "-t",
    help="The template to use for the requests.",
)
@click.option(
    "--system",
    "-s",
    help="The system to use for the requests.",
)
@click.option("--emoji", is_flag=True, help="Add emotions and emojis.")
@click.option(
    "--temperature",
    callback=validate_temperature,
    default=1,
    type=float,
    help="The temperature to use for the requests.",
    show_default=True,
)
@click.option(
    "--tokens",
    is_flag=True,
    help=("Count the number of tokens in the prompt, and display the cost of the request."),
)
@click.option(
    "--no-stream",
    is_flag=True,
    default=False,
    help="Disable the streaming of the response.",
)
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    help="Disable colors and formatting, and print the raw response.",
)
@click.option(
    "--rich",
    "-R",
    is_flag=True,
    default=False,
    help="Force Rich formatting.",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Print debug information.",
)
@click.pass_context
def prompt(
    ctx,
    model,
    template,
    system,
    emoji,
    temperature,
    tokens,
    no_stream,
    raw,
    rich,
    prompt_input,
    debug,
):
    """
    Talk to ChatGPT.

    Example: lmt prompt "Say hello" --emoji
    """
    prompt_input = " ".join(prompt_input).strip()

    # Allow for the appending of an additional prompt to the piped stdin content
    if not sys.stdin.isatty() and prompt_input:
        prompt_input = sys.stdin.read().strip() + "\n___\n" + prompt_input

    if not prompt_input:
        # Read piped or redirected stdin content.
        if not sys.stdin.isatty():
            prompt_input = sys.stdin.read().strip()

        # Allow for structured prompts in the terminal.
        if sys.stdin.isatty():
            input_prompt_instructions = (
                "Write or paste your message below. Use <Enter> for new lines."
                "\nTo send your message, press Ctrl+D."
            )

            if sys.stdout.isatty():
                click.echo(click.style(input_prompt_instructions, fg="yellow"))
                click.echo("---")

            # Display instructions in the terminal only, not in redirected or piped output.
            # This ensures the user sees the instructions without affecting the file output.
            if not sys.stdout.isatty():
                with open("/dev/tty", "w", encoding="UTF-8") as output_stream:
                    click.echo(
                        click.style(input_prompt_instructions, fg="yellow"),
                        file=output_stream,
                    )
                    click.echo("---", file=output_stream)

            # Read user input from stdin
            prompt_input = sys.stdin.read().strip()

    if system and template:
        raise click.BadOptionUsage(
            option_name="template",
            message=click.style(
                "You cannot use both `--template` and `--system` at the same time.",
                fg="red",
            ),
        )

    # If *not* in an interactive shell or redirecting to a file,
    # enable the `--raw` option, viz. disabling `Rich` formatting
    if not sys.stdout.isatty():
        raw = True

    # If `--rich` is enabled, force `--raw` to be disabled
    if rich:
        raw = False

    # If in an interactive shell, add a new line after the prompt for better readability
    if sys.stdout.isatty():
        click.echo()

    prepare_and_generate_response(
        system,
        template,
        model,
        emoji,
        prompt_input,
        temperature,
        tokens,
        no_stream,
        raw,
        debug,
    )

    # Same as above (readibility), but after the LLM's response
    if sys.stdout.isatty():
        click.echo()


@lmt.command()
def models():
    """
    List the available models.
    """
    for model, aliases in VALID_MODELS.items():
        click.echo(model)
        if aliases:
            if len(aliases) == 1:
                click.echo(f"  Alias: {aliases[0]}")
            click.echo(f"  Aliases: {', '.join(aliases)}")


@lmt.group()
def templates():
    """
    Manage the templates.
    """


@templates.command("list")
def print_templates_list():
    """
    List the available templates.
    """
    templates_names_list = sorted([template.stem for template in TEMPLATES_DIR.iterdir()])
    if templates_names_list:
        click.echo("\n".join(templates_names_list))


@templates.command("view")
@click.argument("template")
def view_template(template):
    """
    View a template.
    """
    template = TEMPLATES_DIR / f"{template}.yaml"
    if template.exists():
        with open(template, "r", encoding="UTF-8") as template_file:
            click.echo(template_file.read())


@templates.command()
@click.argument("template")
def edit(template):
    """
    Edit a template.
    """
    template_file = TEMPLATES_DIR / f"{template}.yaml"
    if template_file.exists():
        original_file_content = template_file.read_text()
        click.edit(filename=template_file)

        if original_file_content == template_file.read_text():
            click.echo("No changes were made.")
        else:
            click.echo(
                f"{click.style('Success!', fg='green')} Template"
                f" {click.style(template, fg='green')} was updated."
            )

    else:
        click.echo(
            click.style("Error: ", fg="red")
            + f"Template {click.style(template, fg='red')} does not exist."
        )
        click.echo(f"Use `{click.style(f'lmt templates add {template}', fg='blue')}` to create it.")


@templates.command("add")
@click.argument("template", required=False)
def add_template(template):
    """
    Create a new template
    """
    if not template:
        template = click.prompt("Template name")
        if template in [template.name for template in TEMPLATES_DIR.iterdir()]:
            click.echo(
                click.style("Error: ", fg="red")
                + f"Template {click.style(template, fg='red')} already exists."
            )
            click.echo(
                f"Use `{click.style(f'lmt templates edit {template}', fg='blue')}` to edit it."
            )
            return

    template_file = TEMPLATES_DIR / f"{template}.yaml"
    default_template_file = get_default_template_file_path()

    shutil.copyfile(default_template_file, template_file)

    click.edit(filename=template_file)

    if filecmp.cmp(default_template_file, template_file, shallow=False):
        click.echo(
            click.style("Aborting: ", fg="red")
            + "The template has not been created because no changes were made."
        )
        template_file.unlink()
    else:
        click.echo(
            f"{click.style('Success!', fg='green')} Template"
            f" '{click.style(template, fg='green')}' created."
        )


@templates.command("delete")
@click.argument("template", required=True)
def delete_template(template):
    """
    Delete the template.
    """
    template_file = TEMPLATES_DIR / f"{template}.yaml"
    if template_file.exists():
        click.confirm(
            f"Are you sure you want to delete the template '{template}'?",
            abort=True,
        )
        template_file.unlink()
        click.echo(
            f"{click.style('Success!', fg='green')} Template"
            f" '{click.style(template, fg='blue')}' deleted."
        )
    else:
        click.echo(
            click.style("Error: ", fg="red")
            + f"The template '{click.style(template, fg='red')}' does not exist."
        )


@templates.command("rename")
@click.argument("template", required=True)
def rename_template(template):
    """
    Rename the template.
    """
    template_file = TEMPLATES_DIR / template
    if template_file.exists():
        new_template_name = click.prompt("New template name", default=template)
        new_template_file = TEMPLATES_DIR / new_template_name
        template_file.rename(new_template_file)
        click.echo(
            f"{click.style('Success!', fg='green')} Template"
            f" '{click.style(template, fg='blue')}' renamed to"
            f" '{click.style(new_template_name, fg='green')}'."
        )
    else:
        click.echo(click.style("Error: ", fg="red") + f"The template '{template}' does not exist.")


@lmt.group()
def key():
    """
    Manage the OpenAI API key.
    """


@key.command(name="edit")
def edit_api_key():
    """
    Edit the OpenAI API key.
    """
    edit_key()


@key.command(name="set")
def set_api_key():
    """
    Set the OpenAI API key.
    """
    set_key()



from setuptools import find_packages, setup

VERSION = "0.0.41"


def read_requirements():
    with open("requirements.txt") as file:
        return list(file)


def get_long_description():
    with open("README.md", encoding="utf8") as file:
        return file.read()


setup(
    name="LMterminal",
    description="Interact with OpenAI's models from the comfort of your terminal.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Sébastien De Revière",
    url="https://github.com/sderev/lmterminal",
    project_urls={
        "Documentation": "https://github.com/sderev/lmterminal",
        "Issues": "http://github.com/sderev/lmterminal/issues",
        "Changelog": "https://github.com/sderev/lmterminal/releases",
    },
    license="Apache Licence, Version 2.0",
    version=VERSION,
    packages=find_packages(),
    install_requires=read_requirements(),
    entry_points={
        "console_scripts": [
            "lmt=lmterminal.cli:lmt",
        ]
    },
    python_requires=">=3.8",
)



import os
import sys
import time

import openai
import tiktoken

BLUE = "\x1b[34m"
RED = "\x1b[91m"
RESET = "\x1b[0m"

DEFAULT_MODEL = "gpt-4o-mini"


def format_prompt(system_content, user_content):
    """Returns a formatted prompt for the OpenAI API."""
    return [
        {
            "role": "system",
            "content": system_content,
        },
        {
            "role": "user",
            "content": user_content,
        },
    ]


def chatgpt_request(
    api_key,
    prompt,
    model=DEFAULT_MODEL,
    # max_tokens=3900,
    n=1,
    temperature=1,
    stop=None,
    stream=False,
    update_markdown_stream=None,
):
    """
    Sends a request to the OpenAI Chat API.
    """
    start_time = time.monotonic_ns()
    openai.api_key = os.getenv("OPENAI_API_KEY")

    # Make the API request
    response = openai.ChatCompletion.create(
        api_key=api_key,
        messages=prompt,
        model=model,
        # max_tokens=max_tokens,
        n=n,
        temperature=temperature,
        # stop=stop,
        stream=stream,
    )

    if stream:
        # Create variables to collect the stream of chunks
        collected_chunks = []
        collected_messages = []

        # Iterate through the stream of events
        for chunk in response:
            collected_chunks.append(chunk)  # save the event response
            chunk_message = chunk["choices"][0]["delta"]  # extract the message
            if chunk_message is not None:
                collected_messages.append(chunk_message)  # save the message

            if update_markdown_stream:
                update_markdown_stream(chunk_message.get("content", ""))
            else:
                print(chunk_message.get("content", ""), end="")
        response = collected_chunks

        # Save the time delay and text received
        response_time = (time.monotonic_ns() - start_time) / 1e9
        generated_text = "".join([m.get("content", "") for m in collected_messages])

    else:
        # Extract and save the generated response
        generated_text = response["choices"][0]["message"]["content"]

        # Save the time delay
        response_time = (time.monotonic_ns() - start_time) / 1e9

    return (
        generated_text,
        response_time,
        response,
    )


def num_tokens_from_string(string, model=DEFAULT_MODEL):
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.encoding_for_model(model)
    num_tokens = len(encoding.encode(string))
    return num_tokens


def num_tokens_from_messages(messages, model=DEFAULT_MODEL):
    """Returns the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    tokens_per_message = 3
    tokens_per_name = 1

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def estimated_cost(num_tokens, price_per_1M_tokens):
    """Returns the estimated cost of a number of tokens."""
    return f"{num_tokens / 10**6 * price_per_1M_tokens:.6f}"


def estimate_prompt_cost(message, model):
    """Returns the estimated cost of a prompt."""
    num_tokens = num_tokens_from_messages(message, model)

    # Prices in USD per 1M input tokens
    prices = {
        "gpt-3.5-turbo": 0.50,
        "gpt-3.5-turbo-0125": 0.50,
        "gpt-3.5-turbo-1106": 0.50,
        "gpt-3.5-turbo-instruct": 1.50,
        "gpt-4": 30,
        "gpt-4-turbo-preview": 10,
        "gpt-4-turbo": 10,
        "gpt-4-turbo-2024-04-09": 0.01,
        "gpt-4-0613": 0.03,
        "gpt-4-1106-preview": 10,
        "gpt-4-0125-preview": 10,
        "gpt-4-32k": 60,
        "gpt-4-32k-0613": 60,
        "gpt-4o": 2.50,
        "gpt-4o-2024-05-13": 5,
        "gpt-4o-2024-08-06": 2.50,
        "gpt-4o-2024-11-20": 2.50,
        "gpt-4o-mini": 0.15,
        "gpt-4o-mini-2024-07-18": 0.15,
        "chatgpt-4o-latest": 5,
        "o1": 15,
        "o1-2024-12-17": 15,
        "o1-preview": 15,
        "o1-preview-2024-09-12": 15,
        "o1-mini": 3,
        "o1-mini-2024-09-12": 3,
        "gpt-4.1": 2,
        "gpt-4.1-2025-04-14": 2,
        "gpt-4.1-mini": 0.40,
        "gpt-4.1-mini-2025-04-14": 0.40,
        "gpt-4.1-nano": 0.1,
        "gpt-4.1-nano-2025-04-14": 0.10,
        "gpt-4.5-preview": 75,
        "o3": 10,
        "o3-2025-04-16": 10,
        "o3-mini": 1.10,
        "o3-mini-2025-01-31": 1.10,
        "o4-mini": 1.10,
        "o4-mini-2025-04-16": 1.10,
    }

    return estimated_cost(num_tokens, prices[model])


def handle_rate_limit_error():
    """
    Provides guidance on how to handle a rate limit error.
    """
    sys.stderr.write("\n")
    sys.stderr.write(
        BLUE
        + "You might not have set a usage rate limit in your OpenAI account settings. "
        + RESET
        + "\n"
    )
    sys.stderr.write(
        "If that's the case, you can set it"
        " here:\nhttps://platform.openai.com/account/billing/limits" + "\n"
    )

    sys.stderr.write("\n")
    sys.stderr.write(
        BLUE + "If you have set a usage rate limit, please try the following steps:" + RESET + "\n"
    )
    sys.stderr.write("- Wait a few seconds before trying again.\n")
    sys.stderr.write("\n")
    sys.stderr.write(
        "- Reduce your request rate or batch tokens. You can read the"
        " OpenAI rate limits"
        " here:\nhttps://platform.openai.com/account/rate-limits" + "\n"
    )
    sys.stderr.write("\n")
    sys.stderr.write(
        "- If you are using the free plan, you can upgrade to the paid"
        " plan"
        " here:\nhttps://platform.openai.com/account/billing/overview" + "\n"
    )
    sys.stderr.write("\n")
    sys.stderr.write(
        "- If you are using the paid plan, you can increase your usage"
        " rate limit"
        " here:\nhttps://platform.openai.com/account/billing/limits" + "\n"
    )


def handle_authentication_error():
    """
    Provides guidance on how to handle an authentication error.
    """
    sys.stderr.write(
        f"{RED}Error:{RESET} Your API key or token is invalid, expired, or"
        " revoked. Check your API key or token and make sure it is correct"
        " and active.\n"
    )
    sys.stderr.write(
        "\nYou may need to generate a new API key from your account"
        " dashboard: https://platform.openai.com/account/api-keys\n"
    )
